@@ -1,5 +1,5 @@
 /*
- * Copyright 2001 David Mansfield and Cobite, Inc.
+ * Copyright 2001, 2002, 2003 David Mansfield and Cobite, Inc.
  * See COPYING file for license information 
  */
 
@@ -11,7 +11,7 @@
 #include "hash.h"
 #include "rcsid.h"
 
-RCSID("$Id: hash.c,v 1.5 2001/10/25 18:36:10 adam Exp $");
+RCSID("$Id: hash.c,v 1.6 2003/05/07 15:42:38 david Exp $");
 
 #define HASH_CONST 37
 
@@ -73,42 +73,8 @@ void destroy_hash_table(struct hash_table *tbl, void (*delete_obj)(void *))
  */
 void *put_hash_object(struct hash_table *tbl, const char *key, void *obj)
 {
-    struct list_head *head;
-    struct hash_entry *entry;
-    unsigned int hash;
-    void *retval;
-
-    /* FIXME: how can get_hash_entry be changed to be usable here? 
-     * we need the value of head later if the entry is not found...
-     */
-    hash  = hash_string(key) % tbl->ht_size;
-    head  = &tbl->ht_lists[hash];
-    entry = scan_list(head, key);
-
-    if (entry)
-    {
-	retval = entry->he_obj;
-	entry->he_obj = obj;
-    }
-    else
-    {
-	retval = NULL;
-	entry = (struct hash_entry *)malloc(sizeof(*entry) + strlen(key) + 1);
-	
-	if (!entry)
-	{
-	    debug(DEBUG_APPERROR,"malloc failed put_hash_object key='%s'",key);
-	}
-	else
-	{
-	    entry->he_key = (char *)(entry + 1);
-	    entry->he_obj = obj;
-	    strcpy(entry->he_key, key);
-
-	    list_add(&entry->he_list, head);
-	}
-    }
-
+    void * retval;
+    put_hash_object_ex(tbl, key, obj, HT_KEYCOPY, NULL, &retval);
     return retval;
 }
 
@@ -204,3 +170,103 @@ struct hash_entry *next_hash_entry(struct hash_table *tbl)
 
     return( NULL );
 }
+
+int put_hash_object_ex(struct hash_table *tbl, const char *key, void *obj, int copy, 
+		       char ** oldkey, void ** oldobj)
+{
+    struct list_head *head;
+    struct hash_entry *entry;
+    unsigned int hash;
+    int retval = 0;
+
+    /* FIXME: how can get_hash_entry be changed to be usable here? 
+     * we need the value of head later if the entry is not found...
+     */
+    hash  = hash_string(key) % tbl->ht_size;
+    head  = &tbl->ht_lists[hash];
+    entry = scan_list(head, key);
+
+    if (entry)
+    {
+	if (oldkey)
+	    *oldkey = entry->he_key;
+	if (oldobj)
+	    *oldobj = entry->he_obj;
+
+	/* if 'copy' is set, then we already have an exact
+	 * private copy of the key (by definition of having
+	 * found the match in scan_list) so we do nothing.
+	 * if !copy, then we can simply assign the new
+	 * key
+	 */
+	if (!copy)
+	    entry->he_key = (char*)key; /* discard the const */
+	entry->he_obj = obj;
+    }
+    else
+    {
+	size_t s = sizeof(*entry);
+
+	if (oldkey)
+	    *oldkey = NULL;
+	if (oldobj)
+	    *oldobj = NULL;
+
+	if (copy)
+	    s +=  strlen(key) + 1;
+
+	entry = (struct hash_entry *)malloc(s);
+	
+	if (!entry)
+	{
+	    debug(DEBUG_APPERROR,"malloc failed put_hash_object key='%s'",key);
+	    retval = -1;
+	}
+	else
+	{
+	    if (copy)
+	    {
+		entry->he_key = (char *)(entry + 1);
+		strcpy(entry->he_key, key);
+	    }
+	    else
+	    {
+		entry->he_key = (char*)key; /* discard the const */
+	    }
+
+	    entry->he_obj = obj;
+
+	    list_add(&entry->he_list, head);
+	}
+    }
+
+    return retval;
+}
+
+void destroy_hash_table_ex(struct hash_table *tbl, 
+			   void (*delete_entry)(const void *, char *, void *), 
+			   const void * cookie)
+{
+    struct list_head  *head, *next, *tmp;
+    struct hash_entry *entry;
+    int i;
+    
+    for (i = 0; i < tbl->ht_size; i++)
+    {
+	head = &tbl->ht_lists[i];
+	next = head->next;
+	
+	while (next != head)
+	{
+	    tmp = next->next;
+	    entry = list_entry(next, struct hash_entry, he_list);
+	    if (delete_entry)
+		delete_entry(cookie, entry->he_key, entry->he_obj);
+	    free(entry);
+
+	    next = tmp;
+	}
+    }
+
+    free(tbl);
+}
