@@ -1,5 +1,5 @@
 /*
- * Copyright 2001 David Mansfield and Cobite, Inc.
+ * Copyright 2001, 2002, 2003 David Mansfield and Cobite, Inc.
  * See COPYING file for license information 
  */
 
@@ -23,6 +23,12 @@ struct hash_table
     struct list_head  *iterator_ptr;
 };
 
+enum
+{
+    HT_NO_KEYCOPY,
+    HT_KEYCOPY
+};
+
 #ifdef __cplusplus
 extern "C" {
 #endif
@@ -33,6 +39,9 @@ void *put_hash_object(struct hash_table *tbl, const char *key, void *obj);
 void *get_hash_object(struct hash_table *tbl, const char *key);
 void *remove_hash_object(struct hash_table *tbl, const char *key);
 
+int put_hash_object_ex(struct hash_table *tbl, const char *key, void *obj, int, char **, void **);
+void destroy_hash_table_ex(struct hash_table *tbl, void (*delete_entry)(const void *, char *, void *), const void *);
+
 void reset_hash_iterator(struct hash_table *tbl);
 struct hash_entry *next_hash_entry(struct hash_table *tbl);
 
